from . import logging


class Context(object):
    """Context object for storing context information of 
    request currently being processed.
    """
    def __init__(self, *args, logger = None, headers = None):
        """Create a new Context object

        Keyword arguments:
        logger: Logger that will be used for logging.
        headers: Headers for the request. 
        """
        self._values = {}
        if logger is None:
            logger = logging.get_logger()
        self._logger = logger
        if headers is None:
            headers = {}
        self._headers = headers

    def set(self, key, value):
        """Set a Context value
        
        Arguments:
        key: Key for the context key-value pair.
        value: Value to be stored.
        """
        self._values[key] = value

    def get(self, key):
        """Get a Context value
        
        Arguments:
        key: Key for the context key-value pair.
        """
        return self._values[key]

    def get_logger(self):
        """Get current logger used by Context."""
        return self._logger
    
    def set_logger(self, logger):
        """Set logger for this Context
        
        Arguments:
        logger: Logger object to be used.
        """
        self._logger = logger

    def get_headers(self):
        """Get headers that are currently stored."""
        return self._headers

    def set_header(self, key, value):
        """Set a request header
        
        Arguments:
        key: Key for the header.
        value: Value for the header.
        """
        self._headers[key] = value



import requests

from . import exceptions
from . import errors

class TwirpClient(object):
    def __init__(self, address, timeout=5):
        self._address = address
        self._timeout = timeout

    def _make_request(self, *args, url, ctx, request, response_obj, **kwargs):
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self._timeout
        headers = ctx.get_headers()
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers
        kwargs['headers']['Content-Type'] = 'application/protobuf'
        try:
            resp = requests.post(url=self._address+url, data=request.SerializeToString(), **kwargs)
            if resp.status_code == 200:
                response = response_obj()
                response.ParseFromString(resp.content)
                return response
            else:
                raise exceptions.TwirpServerException.from_json(resp.json())
            # Todo: handle error
        except requests.exceptions.Timeout as e:
            raise exceptions.TwirpServerException(
                code=errors.Errors.DeadlineExceeded,
                message=e.message,
                meta={"original_exception": e},
            )
        except requests.exceptions.ConnectionError as e:
            raise exceptions.TwirpServerException(
                code=errors.Errors.Unavailable,
                message=e.message,
                meta={"original_exception": e},
            )




try:
    import http.client as httplib
except ImportError:
    import httplib

try:
    import ujson as json
except:
    import json

from . import errors



class TwirpServerException(httplib.HTTPException):
    def __init__(self, *args, code, message, meta={}):
        try:
            self._code = errors.Errors(code)
        except ValueError:
            self._code = errors.Errors.Unknown
        self._message = message
        self._meta = meta
        super(TwirpServerException, self).__init__(message)

    @property
    def code(self):
        if isinstance(self._code, errors.Errors):
            return self._code
        return errors.Errors.Unknown

    @property
    def message(self):
        return self._message

    @property
    def meta(self):
        return self._meta

    def to_dict(self):
        err = {
            "code": self._code.value,
            "msg": self._message,
            "meta": {}
        }
        for k, v in self._meta.items():
            err["meta"][k] = str(v)
        return err

    def to_json_bytes(self):
        return json.dumps(self.to_dict()).encode('utf-8')

    @staticmethod
    def from_json(err_dict):
        return TwirpServerException(
            code=err_dict.get('code', errors.Errors.Unknown),
            message=err_dict.get('msg',''),
            meta=err_dict.get('meta',{}),
        )

def InvalidArgument(*args, argument, error):
    return TwirpServerException(
        code=errors.Errors.InvalidArgument,
        message="{} {}".format(argument, error),
        meta={
            "argument":argument
        }
    )

def RequiredArgument(*args, argument):
    return InvalidArgument(
        argument=argument,
        error="is required"
    )



import os
import logging
import sys

import structlog
from structlog.stdlib import LoggerFactory, add_log_level

_configured = False


def configure(force = False):
    """
    Configures logging & structlog modules

    Keyword Arguments:
    force: Force to reconfigure logging.
    """

    global _configured
    if _configured and not force:
        return

    # Check whether debug flag is set
    debug = os.environ.get('DEBUG_MODE', False)
    # Set appropriate log level
    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    # Set logging config
    logging.basicConfig(
        level = log_level,
        format = "%(message)s",
    )

    # Configure structlog
    structlog.configure(
        logger_factory = LoggerFactory(),
        processors = [
            add_log_level,
            # Add timestamp
            structlog.processors.TimeStamper('iso'),
            # Add stack information
            structlog.processors.StackInfoRenderer(),
            # Set exception field using exec info
            structlog.processors.format_exc_info,
            # Render event_dict as JSON
            structlog.processors.JSONRenderer()
        ]
    )

    _configured = True


def get_logger(**kwargs):
    """
    Get the structlog logger
    """
    # Configure logging modules
    configure()
    # Return structlog
    return structlog.get_logger(**kwargs)



class TwirpHook(object):
    # Called as soon as a request is received, always called
    def request_received(self, *args, ctx):
        pass

    # Called once the request is routed, service name known, only called if request is routable
    def request_routed(self,  *args, ctx):
        pass

    # Called once the response is prepared, not called for error cases
    def response_prepared(self,  *args, ctx):
        pass

    # Called if an error occurs
    def error(self,  *args, ctx, exc):
        pass

    # Called after error is sent, always called
    def response_sent(self,  *args, ctx):
        pass


class ChainHooks(TwirpHook):
    def __init__(self, *hooks):
        for hook in hooks:
            assert isinstance(hook, TwirpHook)
        self._hooks = hooks

    def request_received(self, *args, ctx):
        for hook in self._hooks:
            hook.request_received(ctx=ctx)

    def request_routed(self, *args, ctx):
        for hook in self._hooks:
            hook.request_routed(ctx=ctx)

    def response_prepared(self, *args, ctx):
        for hook in self._hooks:
            hook.response_prepared(ctx=ctx)

    def error(self, *args, ctx, exc):
        for hook in self._hooks:
            hook.error(ctx=ctx, exc=exc)


    def response_sent(self, *args, ctx):
        for hook in self._hooks:
            hook.response_sent(ctx=ctx)


from enum import Enum

class Errors(Enum):
    Canceled = "canceled"
    Unknown = "unknown"
    InvalidArgument = "invalid_argument"
    DeadlineExceeded = "deadline_exceeded"
    NotFound = "not_found"
    BadRoute = "bad_route"
    AlreadyExists = "already_exists"
    PermissionDenied = "permission_denied"
    Unauthenticated = "unauthenticated"
    ResourceExhausted = "resource_exhausted"
    FailedPrecondition = "failed_precondition"
    Aborted = "aborted"
    OutOfRange = "out_of_range"
    Unimplemented = "unimplemented"
    Internal = "internal"
    Unavailable = "unavailable"
    DataLoss = "data_loss"
    Malformed = "malformed"
    NoError = ""

    @staticmethod
    def get_status_code(code):
        return {
            Errors.Canceled: 408,
            Errors.Unknown: 500,
            Errors.InvalidArgument: 400,
            Errors.Malformed: 400,
            Errors.DeadlineExceeded: 408,
            Errors.NotFound: 404,
            Errors.BadRoute: 404,
            Errors.AlreadyExists: 409,
            Errors.PermissionDenied: 403,
            Errors.Unauthenticated: 401,
            Errors.ResourceExhausted: 429,
            Errors.FailedPrecondition: 412,
            Errors.Aborted: 409,
            Errors.OutOfRange: 400,
            Errors.Unimplemented: 501,
            Errors.Internal: 500,
            Errors.Unavailable: 503,
            Errors.DataLoss: 500,
            Errors.NoError: 200,
        }.get(code, 500)




from . import exceptions
from . import errors


class TwirpServer(object):

    def __init__(self, *args, service):
        self.service = service
        self._endpoints = {}
        self._prefix = ""

    @property
    def prefix(self):
        return self._prefix

    def get_endpoint(self, path):
        (_, url_pre, rpc_method) = path.rpartition(self._prefix + "/")
        if not url_pre or not rpc_method:
            raise exceptions.TwirpServerException(
                code=errors.Errors.BadRoute,
                message="no handler for path " + path,
                meta={"twirp_invalid_route": "POST " + path},
            )

        endpoint = self._endpoints.get(rpc_method, None)
        if not endpoint:
            raise exceptions.TwirpServerException(
                code=errors.Errors.Unimplemented,
                message="service has no endpoint " + rpc_method,
                meta={"twirp_invalide_route": "POST " + path})


        return endpoint




SERVICE_NAME = 'service_name'
METHOD_NAME = 'method_name'

RAW_HEADERS = 'raw_headers'
RAW_REQUEST_PATH = 'raw_request_path'

RESPONSE_STATUS = 'response_status'

ORIGINAL_EXCEPTION = 'original_exception'



import functools
from collections import namedtuple

from google.protobuf import json_format
from google.protobuf import symbol_database as _symbol_database


from . import context

from . import server
from . import exceptions
from . import errors
from . import hook as vtwirp_hook

_sym_lookup = _symbol_database.Default().GetSymbol

Endpoint = namedtuple("Endpoint", ["service_name", "name", "function", "input", "output"])


class TwirpBaseApp(object):
    def __init__(self, *middlewares, hook=None, prefix="", max_receive_message_length=1024*100*100, ctx_class=None):
        self._prefix = prefix
        self._services = {}
        self._max_receive_message_length = max_receive_message_length
        if ctx_class is None:
            ctx_class = context.Context
        assert issubclass(ctx_class, context.Context)
        self._ctx_class = ctx_class
        self._middlewares = middlewares
        if hook is None:
            hook = vtwirp_hook.TwirpHook()
        assert isinstance(hook, vtwirp_hook.TwirpHook)
        self._hook = hook

    def add_service(self, svc: server.TwirpServer):
        self._services[self._prefix+svc.prefix] = svc

    def _get_endpoint(self, path):
        svc = self._services.get(path.rsplit("/", 1)[0], None)
        if svc is None:
            raise exceptions.TwirpServerException(
                code=errors.Errors.NotFound,
                message="not found"
            )

        return svc.get_endpoint(path[len(self._prefix):])

    @staticmethod
    def json_decoder(body, data_obj=None):
        data = data_obj()
        json_format.Parse(body, data)
        return data

    @staticmethod
    def json_encoder(value, data_obj=None):
        if not isinstance(value, data_obj):
            raise exceptions.TwirpServerException(
                code=errors.Errors.Internal,
                message=("bad service response type " + str(type(value)) +
                 ", expecting: " + data_obj.DESCRIPTOR.full_name))

        return json_format.MessageToJson(value, preserving_proto_field_name=True).encode('utf-8'), {"Content-Type": "application/json"}

    @staticmethod
    def proto_decoder(body, data_obj=None):
        data = data_obj()
        data.ParseFromString(body)
        return data

    @staticmethod
    def proto_encoder(value, data_obj=None):
        if not isinstance(value, data_obj):
            raise exceptions.TwirpServerException(
                code=errors.Errors.Internal,
                message=("bad service response type " + str(type(value)) +
                 ", expecting: " + data_obj.DESCRIPTOR.full_name))

        return value.SerializeToString(), {"Content-Type": "application/protobuf"}

    def _get_encoder_decoder(self, endpoint, headers):
        ctype = headers.get('content-type', None)
        if "application/json" == ctype:
            decoder = functools.partial(self.json_decoder, data_obj=endpoint.input)
            encoder = functools.partial(self.json_encoder, data_obj=endpoint.output)
        elif "application/protobuf" == ctype:
            decoder = functools.partial(self.proto_decoder, data_obj=endpoint.input)
            encoder = functools.partial(self.proto_encoder, data_obj=endpoint.output)
        else:
            raise exceptions.TwirpServerException(
                code=errors.Errors.BadRoute,
                message="unexpected Content-Type: " + ctype
            )
        return encoder, decoder


from setuptools import setup

with open('version.txt') as f:
    version = f.read().strip()

with open('README.md', encoding='utf-8') as f:
      long_description = f.read().strip()

setup(name="twirp",
      version=version,
      description="Twirp server and client lib",
      long_description=long_description,
      long_description_content_type='text/markdown',
      licesnse='unlicense',
      packages=['twirp'],
      install_requires=[
            'requests',
            'structlog',
            'protobuf'
      ],
      test_requires=[
      ],
      zip_safe=False)



import asyncio
import functools
import typing
import traceback

try:
    import ujson as json
except:
    import json

from . import base
from . import exceptions
from . import errors
from . import ctxkeys

try:
    import contextvars  # Python 3.7+ only.
except ImportError:  # pragma: no cover
    contextvars = None  # type: ignore

# Lifted from starlette.concurrency
async def run_in_threadpool(func: typing.Callable, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
    loop = asyncio.get_event_loop()
    if contextvars is not None:  # pragma: no cover
        # Ensure we run in the same context
        child = functools.partial(func, *args, **kwargs)
        context = contextvars.copy_context()
        func = context.run
        args = (child,)
    elif kwargs:  # pragma: no cover
        # loop.run_in_executor doesn't accept 'kwargs', so bind them in here
        func = functools.partial(func, **kwargs)
    return await loop.run_in_executor(None, func, *args)

def thread_pool_runner(func):
    async def run(ctx, request):
        return await run_in_threadpool(func, ctx, request)
    return run

class TwirpASGIApp(base.TwirpBaseApp):

    async def __call__(self, scope, receive, send):
        assert scope['type'] == 'http'
        ctx = self._ctx_class()
        try:
            http_method = scope['method']
            if http_method != "POST":
                raise exceptions.TwirpServerException(
                code=errors.Errors.BadRoute,
                message="unsupported method " + http_method + " (only POST is allowed)",
                meta={"twirp_invalid_route": http_method + " " + scope['path']},
            )

            headers = {k.decode('utf-8'): v.decode('utf-8') for (k,v) in scope['headers']}
            ctx.set(ctxkeys.RAW_REQUEST_PATH, scope['path'])
            ctx.set(ctxkeys.RAW_HEADERS, headers)
            self._hook.request_received(ctx=ctx)

            endpoint = self._get_endpoint(scope['path'])
            headers = {k.decode('utf-8'): v.decode('utf-8') for (k,v) in scope['headers']}
            self.validate_content_length(headers=headers)
            encoder, decoder = self._get_encoder_decoder(endpoint, headers)

            # add headers from request into context
            ctx.set(ctxkeys.SERVICE_NAME, endpoint.service_name)
            ctx.set(ctxkeys.METHOD_NAME, endpoint.name)
            ctx.set(ctxkeys.RESPONSE_STATUS, 200)
            self._hook.request_routed(ctx=ctx)
            raw_receive = await self._recv_all(receive)
            request = decoder(raw_receive)
            response_data = await self._with_middlewares(func=endpoint.function, ctx=ctx, request=request)
            self._hook.response_prepared(ctx=ctx)

            body_bytes, headers = encoder(response_data)
            # Todo: middleware
            await self._respond(
                send=send,
                status=200,
                headers=headers,
                body_bytes=body_bytes
            )
            self._hook.response_sent(ctx=ctx)
        except Exception as e:
            await self.handle_error(ctx, e, scope, receive, send)

    def _with_middlewares(self, *args, func, ctx, request):
        chain = iter(self._middlewares + (func,))
        def bind(fn):
            if not asyncio.iscoroutinefunction(fn):
                fn = thread_pool_runner(fn)
            async def nxt(ctx, request):
                try:
                    cur = next(chain)
                    return await fn(ctx, request, bind(cur))
                except StopIteration:
                    pass
                return await fn(ctx, request)
            return nxt
        return bind(next(chain))(ctx,request)


    async def handle_error(self, ctx, exc, scope, receive, send):
        status = 500
        body_bytes = b'{}'
        logger = ctx.get_logger()
        error_data = {}
        ctx.set(ctxkeys.ORIGINAL_EXCEPTION, exc)
        try:
            if not isinstance(exc, exceptions.TwirpServerException):
                error_data['raw_error'] = str(exc)
                error_data['raw_trace'] = traceback.format_exc()
                logger.exception("got non-twirp exception while processing request",**error_data)
                exc = exceptions.TwirpServerException(
                    code=errors.Errors.Internal,
                    message="Internal non-Twirp Error"
                )

            body_bytes = exc.to_json_bytes()
            status = errors.Errors.get_status_code(exc.code)
        except Exception as e:
            exc = exceptions.TwirpServerException(
                    code=errors.Errors.Internal,
                    message="There was an error but it could not be serialized into JSON"
            )
            error_data['raw_error'] = str(exc)
            error_data['raw_trace'] = traceback.format_exc()
            logger.exception("got exception while processing request",**error_data)
            body_bytes = exc.to_json_bytes()

        ctx.set_logger(logger.bind(**error_data))
        ctx.set(ctxkeys.RESPONSE_STATUS, status)
        self._hook.error(ctx=ctx, exc=exc)
        await self._respond(
            send=send,
            status=status,
            headers={'Content-Type':'application/json'},
            body_bytes=body_bytes
            )
        self._hook.response_sent(ctx=ctx)

    async def _respond(self, *args, send, status, headers, body_bytes):
        headers['Content-Length'] = str(len(body_bytes))
        resp_headers = [(k.encode('utf-8'),v.encode('utf-8')) for (k,v) in headers.items()]
        await send({
                'type': 'http.response.start',
                'status': status,
                'headers': resp_headers,
        })
        await send({
             'type': 'http.response.body',
             'body': body_bytes,
        })

    async def _recv_all(self, receive):
        body = b''
        more_body = True
        while more_body:
            message = await receive()
            body += message.get('body', b'')
            more_body = message.get('more_body', False)

            # the body length exceeded than the size set, raise a valid exception
            # so that proper error is returned to the client
            if self._max_receive_message_length < len(body):
                raise exceptions.TwirpServerException(
                    code=errors.Errors.InvalidArgument,
                    message=F"message body exceeds the specified length of {self._max_receive_message_length} bytes"
                )

        return body

    # we will check content-length header value and make sure it is
    # below the limit set
    def validate_content_length(self, headers):
        try:
            content_length = int(headers.get('content-length'))
        except (ValueError, TypeError):
            return

        if self._max_receive_message_length < content_length:
            raise exceptions.TwirpServerException(
                code=errors.Errors.InvalidArgument,
                message=F"message body exceeds the specified length of {self._max_receive_message_length} bytes"
            )
